/*------------------------------------------------------------------------------
* pos2kml.c : convert positions to google earth kml file
*
*          Copyright (C) 2007 by T.TAKASU, All rights reserved.
*
* version : $Revision: 1.1 $ $Date: 2007/01/22 08:05:17 $
* history : 2007/01/20  1.0 new
*-----------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

static const char rcsid[]="$Id: pos2kml.c,v 1.1 2007/01/22 08:05:17 ttaka Exp $";

/* help text -----------------------------------------------------------------*/
static const char *help[]={
"",
" usage: pos2kml [option]... file [...]",
"",
" Read position file(s) and convert it to Google Earth KML file. Each line in",
" the input file shall contain fields of time, position fields (latitude/",
" longitude/height or x/y/z-ecef), and quality flag(option). The line started",
" with '%', '#', ';' is treated as comment. Command options are as follows.",
" ([]:default)",
"",
" -h        print help",
" -o file   output file [infile + .kml]",
" -c color  track color (0:off,1:white,2:red,3:orange,4:green,5:yellow) [5]",
" -p color  point color (0:off,1:white,2:red,3:orange,4:green,5:by qflag) [5]",
" -a        output altitude information [off]",
" -f n e h  add north/east/height offset to position (m) [0 0 0]",
" -e        input x/y/z-ecef position [latitude/longitude/height]",
" -g        input latitude/longitude in the form of ddd mm ss.ss' [ddd.ddd]",
" -s sep    field separator [' ']"
};

#define POSMAX  100000          /* max number of position records */
#define SIZP    0.2             /* mark size of rover positions */
#define SIZR    0.4             /* mark size of reference position */
#define REFSTR  "ref pos"       /* reference position string in comment line */

#define PI      3.1415926535897932 /* pi */
#define D2R     (PI/180.0)      /* deg to rad */
#define R2D     (180.0/PI)      /* rad to deg */
#define RE      6378137.0       /* earth semimajor axis (WGS84) (m) */
#define FE      (1.0/298.257223563) /* earth flattening (WGS84) */

/* print help ----------------------------------------------------------------*/
static void printhelp(void)
{
    int i;
    for (i=0;i<sizeof(help)/sizeof(*help);i++) fprintf(stderr,"%s\n",help[i]);
    exit(0);
}
/* transform ecef to geodetic postion ----------------------------------------*/
static void ecef2pos(const double *r, double *pos)
{
    double e2=FE*(2.0-FE),r2=r[0]*r[0]+r[1]*r[1],z,zk,v=RE,sinp;
    for (z=r[2],zk=0.0;fabs(z-zk)>=1E-4;) {
        zk=z;
        sinp=z/sqrt(r2+z*z);
        v=RE/sqrt(1.0-e2*sinp*sinp);
        z=r[2]+v*e2*sinp;
    }
    pos[0]=r2>1E-12?atan(z/sqrt(r2)):(r[2]>0.0?PI/2.0:-PI/2.0);
    pos[1]=atan2(r[1],r[0]);
    pos[2]=sqrt(r2+z*z)-v;
}
/* output track --------------------------------------------------------------*/
static void outtrack(FILE *f, const double *pos, int n, const double *mpos,
                     const char *color, int outalt)
{
    int i;
    
    fprintf(f,"<Placemark>\n");
    fprintf(f,"  <name>Rover Track</name>\n");
    fprintf(f,"  <LookAt>\n");
    fprintf(f,"    <longitude>%14.9f</longitude>\n",mpos[1]);
    fprintf(f,"    <latitude>%14.9f</latitude>\n",mpos[0]);
    fprintf(f,"  </LookAt>\n");
    fprintf(f,"  <Style>\n");
    fprintf(f,"    <LineStyle>\n");
    fprintf(f,"      <color>%s</color>\n",color);
    fprintf(f,"    </LineStyle>\n");
    fprintf(f,"  </Style>\n");
    fprintf(f,"  <LineString>\n");
    if (outalt) fprintf(f,"    <altitudeMode>absolute</altitudeMode>\n");
    fprintf(f,"    <coordinates>\n");
    for (i=0;i<n;i++) {
        fprintf(f,"      %14.9f,%14.9f,%6.1f\n",pos[1+i*4],pos[i*4],
                outalt?pos[2+i*4]:0.0);
    }
    fprintf(f,"    </coordinates>\n");
    fprintf(f,"  </LineString>\n");
    fprintf(f,"</Placemark>\n");
}
/* output point --------------------------------------------------------------*/
static void outpoint(FILE *f, const double *pos, const char *label, double size,
                     int color, int outalt)
{
    char *icon,*url="http://maps.google.com/mapfiles/kml/";
    
    switch (color) {
        case 1:  icon="pal2/icon18.png"; break; /* white */
        case 2:  icon="pal4/icon55.png"; break; /* red */
        case 3:  icon="pal3/icon49.png"; break; /* orange */
        default: icon="pal3/icon53.png"; break; /* green */
    }
    fprintf(f,"<Placemark>\n");
    if (*label) fprintf(f,"  <name>%s</name>\n",label);
    fprintf(f,"  <Style>\n");
    fprintf(f,"    <IconStyle>\n");
    fprintf(f,"      <scale>%.1f</scale>\n",size);
    fprintf(f,"      <Icon><href>%s%s</href></Icon>\n",url,icon);
    fprintf(f,"    </IconStyle>\n");
    fprintf(f,"  </Style>\n");
    fprintf(f,"  <Point>\n");
    if (outalt) {
        fprintf(f,"    <extrude>1</extrude>\n");
        fprintf(f,"    <altitudeMode>absolute</altitudeMode>\n");
    }
    fprintf(f,"    <coordinates>%14.9f,%14.9f,%6.1f</coordinates>\n",pos[1],
            pos[0],outalt?pos[2]:0.0);
    fprintf(f,"  </Point>\n");
    fprintf(f,"</Placemark>\n");
}
/* save kml file -------------------------------------------------------------*/
static void savekml(const char *file, const double *pos, int n,
                    const double *rpos, const double *mpos, int tcolor,
                    int pcolor, int outalt)
{
    FILE *f;
    int i;
    char *color[]={"ffffffff","ff0000ff","ff00aaff","ff00ff00","ff00ffff"};
    
    f=fopen(file,"w");
    if (f<0) {
        fprintf(stderr,"file open error : %s\n",file);
        return;
    }
    fprintf(f,"<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
    fprintf(f,"<kml xmlns=\"http://earth.google.com/kml/2.0\">\n");
    fprintf(f,"<Document>\n");
    if (tcolor>0) {
        outtrack(f,pos,n,mpos,color[tcolor-1],outalt);
    }
    if (pcolor>0) {
        fprintf(f,"<Folder>\n");
        fprintf(f,"  <name>Rover Position</name>\n");
        for (i=0;i<n;i++) {
            outpoint(f,pos+i*4,"",SIZP,pcolor==5?(int)pos[3+i*4]+1:pcolor,outalt);
        }
        fprintf(f,"</Folder>\n");
    }
    if (rpos[1]!=0.0&&rpos[3]>=-100.0) {
        outpoint(f,rpos,"Reference Position",SIZR,1,outalt);
    }
    fprintf(f,"</Document>\n");
    fprintf(f,"</kml>\n");
    fclose(f);
}
/* fields to numbers ---------------------------------------------------------*/
static int tonum(char *p, const char *sep, double *v)
{
    int n=0;
    for (p=strtok(p,sep);p;p=strtok(NULL,sep)) v[n++]=atof(p);
    return n;
}
/* deg/min/sec to deg --------------------------------------------------------*/
static double todeg(const double *dms)
{
    return dms[0]+dms[1]/60.0+dms[2]/3600.0;
}
/* convert position file to google earth kml file ----------------------------*/
static void convtokml(const char *infile, const char *outfile, double *offset,
                      int tcolor, int pcolor, int posf, int degf, int outalt,
                      const char *sep)
{
    FILE *fp;
    static double pos[4*POSMAX];
    int i,j,n=0,m,q,nq[6]={0};
    char *p,file[256],buff[256];
    double rpos[3]={0.0},mpos[3]={0.0},v[9];
    
    if (!*outfile) {
        if ((p=strrchr(infile,'.'))) {
            strncpy(file,infile,p-infile);
            strcpy(file+(p-infile),".kml");
        }
        else sprintf(file,"%s.kml",infile);
    }
    else strcpy(file,outfile);
    
    if (!(fp=fopen(infile,"r"))) {
        fprintf(stderr,"infile open error : %s\n",infile);
        return;
    }
    while (fgets(buff,sizeof(buff),fp)&&n<POSMAX) {
        
        if (*buff=='%'||*buff=='#'||*buff==';') { /* comment line */
            if (!strstr(buff,REFSTR)||!(p=strchr(buff,':'))) continue;
            m=tonum(p+1,sep,v);
            if (posf&&m>=3) {
                ecef2pos(v,rpos); rpos[0]*=R2D; rpos[1]*=R2D;
            }
            else if (degf&&m>=7) {
                rpos[0]=todeg(v); rpos[1]=todeg(v+3); rpos[2]=v[6];
            }
            else if (m>=3) {
                for (i=0;i<3;i++) rpos[i]=v[i];
            }
            continue;
        }
        p=(sscanf(buff,"%lf/%lf/%lf",v,v+1,v+2)>=3)?buff+11:buff;
        m=tonum(p,sep,v);
        
        if (posf&&m>=4) {
            ecef2pos(v+1,pos+n*4); pos[n*4]*=R2D; pos[1+n*4]*=R2D;
            if (m>4) pos[3+n*4]=v[4];
        }
        else if (degf&&m>=8) {
            pos[n*4]=todeg(v+1); pos[1+n*4]=todeg(v+4); pos[2+n*4]=v[7];
            if (m>8) pos[3+n*4]=v[8];
        }
        else if (m>=4) {
            for (i=0;i<4;i++) pos[i+n*4]=v[i+1];
            if (m>4) pos[3+n*4]=v[4];
        }
        if ((q=(int)pos[3+n*4])>0) nq[q<6?q-1:5]++;
        if (m>=4) n++;
    }
    fclose(fp);
    
    /* mean position */
    for (i=0;i<3;i++) {
        for (j=0;j<n;j++) mpos[i]+=pos[i+j*4];
        mpos[i]/=n;
    }
    /* add offset */
    offset[0]/=RE*D2R;
    offset[1]/=RE*D2R*cos(mpos[0]*D2R);
    for (i=0;i<n;i++) {
        for (j=0;j<3;j++) pos[j+i*4]+=offset[j];
    }
    if (rpos[0]!=0.0) {
        for (j=0;j<3;j++) rpos[j]+=offset[j];
    }
    /* save kml file */
    savekml(file,pos,n,rpos,mpos,tcolor,pcolor,outalt);
                                    
    /* print counts */
    printf("%s : n=%d (Q=",file,n);
    for (i=1;i<=6;i++)
        printf("%d:%d(%.1f%%)%s",i,nq[i-1],n<=0?0.0:100.0*nq[i-1]/n,i<6?",":"");
    printf(")\n");
}
/* pos2kml main --------------------------------------------------------------*/
int main(int argc, char **argv)
{
    int i,j,n,posf=0,degf=0,outalt=0,tcolor=5,pcolor=5;
    char *infile[32],*outfile="",*sep=" ";
    double offset[3]={0.0};
    
    for (i=1,n=0;i<argc;i++) {
        if      (!strcmp(argv[i],"-o")&&i+1<argc) outfile=argv[++i];
        else if (!strcmp(argv[i],"-c")&&i+1<argc) tcolor=atoi(argv[++i]);
        else if (!strcmp(argv[i],"-p")&&i+1<argc) pcolor=atoi(argv[++i]);
        else if (!strcmp(argv[i],"-s")&&i+1<argc) sep=argv[++i];
        else if (!strcmp(argv[i],"-f")&&i+3<argc) {
            for (j=0;j<3;j++) offset[j]=atof(argv[++i]);
        }
        else if (!strcmp(argv[i],"-e")) posf=1;
        else if (!strcmp(argv[i],"-g")) degf=1;
        else if (!strcmp(argv[i],"-a")) outalt=1;
        else if (*argv[i]=='-') printhelp();
        else if (n<32) infile[n++]=argv[i];
    }
    if (tcolor<0||5<tcolor||pcolor<0||5<pcolor) {
        fprintf(stderr,"%s : command option error\n",__FILE__);
        return -1;
    }
    if (n<=0) {
        fprintf(stderr,"%s : no input file\n",__FILE__);
        return -1;
    }
    for (i=0;i<n;i++) {
        convtokml(infile[i],outfile,offset,tcolor,pcolor,posf,degf,outalt,sep);
    }
    return 0;
}
