/*------------------------------------------------------------------------------
* readrnx.c : read rinex observation data/navigation message files
*
*          Copyright (C) 2007 by T.TAKASU, All rights reserved.
*
* version : $Revision: 1.7 $ $Date: 2007/01/23 16:28:05 $
* history : 2006/01/16 1.0 new
*-----------------------------------------------------------------------------*/
#include "rtklib.h"

static const char rcsid[]="$Id: readrnx.c,v 1.7 2007/01/23 16:28:05 ttaka Exp $";

/* constants -----------------------------------------------------------------*/

const char *obstypes[NFREQ*2]={"L1","L2","C1","P2"}; /* read obs types */

/* sort observation data by time, rcv, sat -----------------------------------*/
static int cmpobs(const void *p1, const void *p2)
{
    obsd_t *q1=(obsd_t *)p1,*q2=(obsd_t *)p2;
    double tt=timediff(q1->time,q2->time);
    
    if (fabs(tt)>MAXDT) return tt<0?-1:1;
    return q1->rcv!=q2->rcv?q1->rcv-q2->rcv:q1->sat-q2->sat;
}
static int sortobs(obs_t *obs)
{
    int i,j,n;
    
    if (obs->n<=0) return 0;
    qsort(obs->data,obs->n,sizeof(obsd_t),cmpobs);
    for (i=n=0;i<obs->n;i=j,n++) {
        for (j=i+1;j<obs->n;j++)
            if (timediff(obs->data[j].time,obs->data[i].time)>MAXDT) break;
    }
    return n;
}
/* sort satellite ephemerides by toe, sat and delete duplicate ---------------*/
static int cmpeph(const void *p1, const void *p2)
{
    eph_t *q1=(eph_t *)p1,*q2=(eph_t *)p2;
    
    return q1->toe.time!=q2->toe.time?(int)(q1->toe.time-q2->toe.time):
          (q1->toe.sec<q2->toe.sec?-1:(q1->toe.sec>q2->toe.sec?1:
           q1->sat-q2->sat));
}
static void uniqeph(nav_t *nav)
{
    int i,j;
    
    if (nav->n<=0) return;
    qsort(nav->eph,nav->n,sizeof(eph_t),cmpeph);
    for (i=j=0;i<nav->n;i++) {
        if (nav->eph[i].sat!=nav->eph[j].sat||
            fabs(timediff(nav->eph[i].toe,nav->eph[j].toe))>0.001)
            nav->eph[++j]=nav->eph[i];
    }
    nav->n=j+1;
}
/* read rinex header ---------------------------------------------------------*/
static char readrnxh(FILE *fp, char tobs[][3], int *nt, double *ion)
{
    char buff[256],trnx=' ';
    int i,j,n;
    
    *nt=0;
    while (fgets(buff,sizeof(buff),fp)) {
        if (strlen(buff)<=60) continue;
        else if (strstr(buff+60,"RINEX VERSION / TYPE")) {
            trnx=*(buff+20);
        }
        else if (strstr(buff+60,"# / TYPES OF OBSERV")) {
            n=(int)str2num(buff,0,6);
            for (i=0,j=10;i<n;i++,j+=6,(*nt)++) {
                if (j>=58) {if (!fgets(buff,sizeof(buff),fp)) break; j=10;}
                strncpy(tobs[i],buff+j,2); tobs[i][2]='\0';
            }
        }
        else if (strstr(buff+60,"ION ALPHA")) {
            for (i=0,j=3;i<4;i++,j+=12) ion[i]=str2num(buff,j,11);
        }
        else if (strstr(buff+60,"ION BETA")) {
            for (i=0,j=3;i<4;i++,j+=12) ion[i+4]=str2num(buff,j,11);
        }
        else if (strstr(buff+60,"END OF HEADER")) break;
    }
    return trnx;
}
/* read rinex observation data -----------------------------------------------*/
static void readrnxo(FILE *fp, char tobs[][3], int nt, int rcv, obs_t *obs,
                     int *osiz)
{
    char buff[256];
    short lock;
    int i,j,k,n=0,index[32]={0},sats[128];
    double value;
    gtime_t time;
    obsd_t *data;
    
    for (i=0;i<nt;i++) {
        for (j=0;j<NFREQ*2;j++) if (!strcmp(tobs[i],obstypes[j])) index[i]=j+1;
    }
    i=-1;
    while (fgets(buff,sizeof(buff),fp)) {
        if (i<0) {
            if (str2time(buff,0,26,&time)) continue;
            n=(int)str2num(buff,29,3);
            for (j=0,k=32;j<n;j++,k+=3) {
                if (k>=68) {if (!fgets(buff,sizeof(buff),fp)) break; k=32;}
                sats[j]=strchr(" G",*(buff+k))?(int)str2num(buff,k+1,2):0;
            }
        }
        else if (sats[i]>=0) {
            if (*osiz<=obs->n) {
                if (*osiz==0) *osiz=32768; else *osiz*=2;
                if (!(obs->data=realloc(obs->data,sizeof(obsd_t)*(*osiz)))) {
                    fprintf(stderr,"%s : memory allocation error\n",__FILE__);
                    *osiz=obs->n=0; return;
                }
            }
            data=&obs->data[obs->n++];
            data->time=time; data->sat=sats[i]; data->rcv=rcv;
            for (j=0;j<2;j++) data->L[j]=data->P[j]=0.0;
            for (j=k=0;j<nt;j++,k+=16) {
                if (k>=80) {if (!fgets(buff,sizeof(buff),fp)) break; k=0;}
                value=str2num(buff,k,14); if (fabs(value)<1E-6) continue;
                lock=(short)str2num(buff,k+14,1)&1; /* 1:loss of lock */
                if (!index[j]) continue;
                if (index[j]<=NFREQ) {
                    data->L[index[j]-1]=value; data->LLI[index[j]-1]=lock;
                }
                else data->P[index[j]-NFREQ-1]=value;
            }
        }
        if (++i>=n) i=-1;
    }
}
/* read rinex navigation message ---------------------------------------------*/
static void readrnxn(FILE *fp, const double *ion, nav_t *nav, int *nsiz)
{
    char buff[256];
    int i=0,j,sat=1;
    double e[64];
    gtime_t toc;
    eph_t *eph;
    
    while (fgets(buff,sizeof(buff),fp)) {
        if (i==0) {
            sat=(int)str2num(buff,0,2);
            if (str2time(buff,3,17,&toc)) continue;
        }
        for (j=i==0?22:3;j<=60;j+=19) e[i++]=str2num(buff,j,19);
        if (i>=31) {
            if (*nsiz<=nav->n) {
                if (*nsiz==0) *nsiz=1024; else *nsiz*=2;
                if (!(nav->eph=realloc(nav->eph,sizeof(eph_t)*(*nsiz)))) {
                    fprintf(stderr,"%s : memory allocation error\n",__FILE__);
                    *nsiz=nav->n=0; return;
                }
            }
            eph=&nav->eph[nav->n++];
            eph->sat=sat; eph->iode=(int)e[3]; eph->iodc=(int)e[26];
            eph->sva=(int)e[23]; eph->svh=(int)e[24];
            eph->toe=gpst2time((int)e[21],e[11]); eph->toc=toc; 
            eph->ttr=gpst2time((int)e[21],e[27]);
            eph->A=e[10]*e[10]; eph->e=e[ 8]; eph->i0 =e[15]; eph->OMG0=e[13];
            eph->omg =e[17]; eph->M0 =e[ 6]; eph->deln=e[ 5]; eph->OMGd=e[18];
            eph->idot=e[19]; eph->crc=e[16]; eph->crs =e[ 4]; eph->cuc =e[ 7];
            eph->cus =e[ 9]; eph->cic=e[12]; eph->cis =e[14]; eph->toes=e[11];
            eph->f0  =e[ 0]; eph->f1 =e[ 1]; eph->f2  =e[ 2]; eph->tgd =e[25];
            i=0;
        }
    }
    for (i=0;i<8;i++) nav->ion[i]=ion[i];
}
/* read rinex file -----------------------------------------------------------*/
static void readrnxf(FILE *fp, obs_t *obs, nav_t *nav, int *rcv, int *osiz,
                     int *nsiz)
{
    int nt;
    char tobs[32][3];
    double ion[8]={ /* default ionosphere model parameters */
        0.1676E-7,-0.7451E-8,-0.5960E-7,0.1192E-6,1.352E5,-1.802E5,6.554E4,0.0
    };
    switch (readrnxh(fp,tobs,&nt,ion)) {
        case 'O': readrnxo(fp,tobs,nt,++(*rcv),obs,osiz); break;
        case 'N': readrnxn(fp,ion,nav,nsiz); break;
    }
}
/* read rinex files ------------------------------------------------------------
* read rinex observation data(obs)/navigation message(nav) files
* args   : char **files  I      rinex files
*          int   n       I      number of rinex files (0:input from stdin)
*          obs_t *obs    O      observation data
*          nav_t *nav    O      navigation messages
* return : number of epochs (0:no data)
* note   : obs data are sorted by time, receiver and satellite number
*          RINEX 2.10, only support gps.
*-----------------------------------------------------------------------------*/
extern int readrnx(char **files, int n, obs_t *obs, nav_t *nav)
{
    FILE *fp;
    int i,rcv=0,osiz=0,nsiz=0;
    
    obs->n=nav->n=0; obs->data=NULL; nav->eph=NULL;
    
    if (n>0) {
        for (i=0;i<n;i++) {
            if (!(fp=fopen(files[i],"r"))) {
                fprintf(stderr,"file open error : %s\n",files[i]);
                continue;
            }
            readrnxf(fp,obs,nav,&rcv,&osiz,&nsiz);
            fclose(fp);
        }
    }
    else readrnxf(stdin,obs,nav,&rcv,&osiz,&nsiz);
    
    if (obs->n>0) obs->data=(obsd_t *)realloc(obs->data,sizeof(obsd_t)*obs->n);
    if (nav->n>0) nav->eph=(eph_t *)realloc(nav->eph,sizeof(eph_t)*nav->n);
    
    uniqeph(nav);
    return sortobs(obs);
}
