/*------------------------------------------------------------------------------
* rnx2rtkp.c : read rinex obs/nav files and compute receiver positions
*
*          Copyright (C) 2007 by T.TAKASU, All rights reserved.
*
* version : $Revision: 1.10 $ $Date: 2007/01/23 16:28:05 $
* history : 2006/01/16  1.0 new
*-----------------------------------------------------------------------------*/
#include "rtklib.h"

static const char rcsid[]="$Id: rnx2rtkp.c,v 1.10 2007/01/23 16:28:05 ttaka Exp $";

static const char prog[]="rnx2rtkp ver.1.0";

/* help text -----------------------------------------------------------------*/
static const char *help[]={
"",
" usage: rnx2rtkp [option]... file file [...]",
"",
" Read RINEX OBS/NAV files, compute receiver (rover) positions and output",
" position solutions. The first RINEX OBS file shall contain receiver (rover)",
" observations. For the relative mode, the second RINEX OBS file shall contain",
" reference (base) receiver observations. At least one RINEX NAV file shall be",
" included in input files. Command options are as follows. ([]:default)",
"",
" -h        print help",
" -o output output file [stdout]",
" -p mode   mode (0:single,1:dgps,2:kinematic,3:static) [2]",
" -m mask   elevation mask angle (deg) [10]",
" -f freq   number of frequencies for relative mode (1:L1,2:L1+L2) [2]",
" -v thres  validation threshold for integer ambiguity (0.0:no AR) [3.0]",
" -b        backward solutions [off]",
" -c        forward/backward combined solutions [off]",
" -i        instantaneous integer ambiguity resolution [off]",
" -e        output x/y/z-ecef position [latitude/longitude/height]",
" -n        output NMEA-0183 GGA sentence [off]",
" -g        output latitude/longitude in the form of ddd mm ss.ss' [ddd.ddd]",
" -t        output time in the form of yyyy/mm/dd hh:mm:ss.ss [sssss.ss]",
" -u        output time in utc [gpst]",
" -d col    columns of time under decimal point [3]",
" -s sep    field separator [' ']",
" -r x y z  reference (base) receiver ecef pos (m) [average of single pos]"
};
/* type ----------------------------------------------------------------------*/
typedef struct {            /* analysis results struct */
    gtime_t t;              /* time */
    sol_t sol,solf;         /* float, dgps or single solution, fixed solution */
    int stat,ns;            /* status, number of satellite */
} result_t;

/* constants/global variables ------------------------------------------------*/

#define MAXFILE     4       /* max number of input files */
#define ELMINR      10.0    /* elevation mask for reference station (deg) */
#define COMMENTH    "%"     /* comment line indicator */

static int mode=2;          /* mode (0:single,1:dgps,2:kinematic,3:static) */
static int nfreq=2;         /* number of frequency (1:L1,2:L1+L2) */
static double elmin=10.0;   /* elevation mask angle (deg) */
static int dirs=0;          /* solutions (0:forward,1:backward,2:combined) */
static int opt=0;           /* instantaneous integer ambiguity resolution */
static double thres=3.0;    /* threshold of integer ambiguity validation */
static int times=0;         /* time system (0:gpst,1:utc) */
static int timef=0;         /* time form (0:sssss.s,1:yyyy/mm/dd hh:mm:ss.s) */
static int timeu=3;         /* columns of time under decimal point */
static int posf=0;          /* position form (0:lat,lon,hgt,1:x,y,z-ecef,2:nmea) */
static int degf=0;          /* latitude/longitude form (0:ddd.ddd,1:ddd mm ss) */
static char *sep=" ";       /* field separator */
static double rb[3]={0.0};  /* reference (base) receiver ecef pos (m) */

static FILE *fp;            /* output file pointer */
static obs_t obss;          /* observation data */
static nav_t navs;          /* navigation messages */
static int nepoch=0;        /* number of observation epochs */
static int iobs=0;          /* current observation data record index */
static int revs=0;          /* analysis direction (0:forward,1:backward) */
static result_t *resf;      /* forward analysis results */
static result_t *resb;      /* backward analysis results */
static int iresf=0;         /* current forward analysis result index */
static int iresb=0;         /* current backward analysis result index */

/* print help ----------------------------------------------------------------*/
static void printhelp(void)
{
    int i;
    for (i=0;i<sizeof(help)/sizeof(*help);i++) fprintf(stderr,"%s\n",help[i]);
    exit(0);
}
/* deg to deg/min/sec --------------------------------------------------------*/
static void todms(double deg, double *dms)
{
    dms[0]=floor(deg);
    dms[1]=floor((deg-dms[0])*60.0);
    dms[2]=(deg-dms[0]-dms[1]/60.0)*3600;
}
/* compute reference station position as average of single position ----------*/
static int refpos(double *rb)
{
    int i,j,k,n;
    double rs[3*MAXOBS],dts[MAXOBS],azel[2*MAXOBS],rr[4],Q[4*4],dtr;
    
    for (i=n=0;i<obss.n;i=j) {
        for (;i<obss.n;i++) if (obss.data[i].rcv==2) break;
        for (j=i;j<obss.n;j++) {
            if (timediff(obss.data[j].time,obss.data[i].time)>0.0) break;
        }
        satpos(obss.data+i,j-i,&navs,rs,dts);
        
        if (pntpos(obss.data+i,j-i,rs,dts,navs.ion,ELMINR*D2R,rr,Q,&dtr,azel)<0)
            continue;
        
        for (k=0;k<3;k++) rb[k]+=rr[k]; n++;
    }
    if (n<=0) return -1; else for (k=0;k<3;k++) rb[k]/=n;
    return 0;
}
/* output header -------------------------------------------------------------*/
static void outhead(char **file, int n)
{
    int i,w1,w2;
    char s1[32],s2[32],*s3[]={"single","dgps","kinematic","static"};
    char *s4[]={"C1","L1","L1+L2"},*s5,*s6,*s7[]={"","/backward","/combined"};
    double t1,t2,pos[3],dms1[3],dms2[3];
    gtime_t ts,te;
    
    fprintf(fp,"%s program   : %s\n",COMMENTH,prog);
    fprintf(fp,"%s inputs    : ",COMMENTH);
    for (i=0;i<n;i++) fprintf(fp,"%s%s",file[i],i<n-1?" ":"");
    fprintf(fp,"\n");
    ts=obss.data[0].time;
    te=obss.data[obss.n-1].time;
    time2str(times?gpst2utc(ts):ts,s1,1);
    time2str(times?gpst2utc(te):te,s2,1);
    t1=time2gpst(ts,&w1);
    t2=time2gpst(te,&w2);
    s5=times?"UTC":"GPST";
    s6=posf!=1?" lat/lon/hight=WGS84/ellipsoidal,":" x/y/z-ecef=WGS84,";
    fprintf(fp,"%s obs start : %s %s (gpsweek%04d %7.1fs)\n",COMMENTH,s1,s5,w1,t1);
    fprintf(fp,"%s obs end   : %s %s (gpsweek%04d %7.1fs)\n",COMMENTH,s2,s5,w2,t2);
    fprintf(fp,"%s mode/obsv : %s/%s%s%s%s\n",COMMENTH,s3[mode],
            s4[!mode?0:(nfreq>1?2:1)],opt?"/instantaneous AR":"",
            thres<=0.0?"/float":"",s7[dirs]);
    fprintf(fp,"%s elev mask : %4.1f deg\n",COMMENTH,elmin);
    if (mode>=1) {
        fprintf(fp,"%s ref pos   :",COMMENTH);
        if (posf==1) {
            fprintf(fp,"%14.4f%s%14.4f%s%14.4f\n",rb[0],sep,rb[1],sep,rb[2]);
        }
        else {
            ecef2pos(rb,pos);
            if (degf) {
                todms(pos[0]*R2D,dms1); todms(pos[1]*R2D,dms2);
                fprintf(fp,"%3.0f%s%02.0f%s%08.5f%s%4.0f%s%02.0f%s%08.5f%s%10.4f\n",
                        dms1[0],sep,dms1[1],sep,dms1[2],sep,dms2[0],sep,dms2[1],
                        sep,dms2[2],sep,pos[2]);
            }
            else {
                fprintf(fp,"%13.9f%s%14.9f%s%10.4f\n",pos[0]*R2D,sep,pos[1]*R2D,
                        sep,pos[2]);
            }
        }
    }
    fprintf(fp,"%s\n%s (time=%s,%s Q=1:fix,2:float,4:dgps,5:single,ns=# of sats)\n",
            COMMENTH,COMMENTH,s5,s6);
    fprintf(fp,"%s  %-*s",COMMENTH,(timef?17:5)+(timeu<=0?0:timeu+1),"time");
    if (posf==1) {
        fprintf(fp,"%14s %14s %14s %3s %3s %8s %8s %8s\n","x-ecef(m)","y-ecef(m)",
                "z-ecef(m)","Q","ns","sdx(m)","sdy(m)","sdz(m)");
    }
    else if (degf) {
        fprintf(fp,"%16s %16s %10s %3s %3s %8s %8s %8s\n","latitude(d'\")",
                "longitude(d'\")","hight(m)","Q","ns","sdn(m)","sde(m)","sdu(m)");
    }
    else {
        fprintf(fp,"%14s %14s %10s %3s %3s %8s %8s %8s\n","latitude(deg)",
                "longitude(deg)","hight(m)","Q","ns","sdn(m)","sde(m)","sdu(m)");
    }
}
/* output result in the form of x/y/z-ecef ----------------------------------*/
static void outecef(FILE *fp, const char *s, double *rr, double *Qr, int stat,
                    int ns)
{
    fprintf(fp,"%s%s%14.4f%s%14.4f%s%14.4f%s%3d%s%3d%s%8.4f%s%8.4f%s%8.4f\n",
            s,sep,rr[0],sep,rr[1],sep,rr[2],sep,stat,sep,ns,sep,sqrt(Qr[0]),sep,
            sqrt(Qr[4]),sep,sqrt(Qr[8]));
}
/* output result in the form of lat/lon/height -------------------------------*/
static void outpos(FILE *fp, const char *s, double *rr, double *Qr, int stat,
                   int ns)
{
    double pos[3],dms1[3],dms2[3],Qe[9];
    
    ecef2pos(rr,pos); covenu(pos,Qr,Qe);
    if (degf) {
        todms(pos[0]*R2D,dms1); todms(pos[1]*R2D,dms2);
        fprintf(fp,"%s%s%4.0f%s%02.0f%s%08.5f%s%4.0f%s%02.0f%s%08.5f",s,sep,
                dms1[0],sep,dms1[1],sep,dms1[2],sep,dms2[0],sep,dms2[1],sep,
                dms2[2]);
    }
    else fprintf(fp,"%s%s%14.9f%s%14.9f",s,sep,pos[0]*R2D,sep,pos[1]*R2D);
    fprintf(fp,"%s%10.4f%s%3d%s%3d%s%8.4f%s%8.4f%s%8.4f\n",sep,pos[2],sep,
            stat,sep,ns,sep,sqrt(Qe[4]),sep,sqrt(Qe[0]),sep,sqrt(Qe[8]));
}
/* output result in the form of NMEA-0183 GGA sentence -----------------------*/
static void outnmea(FILE *fp, gtime_t t, double *rr, int stat, int ns)
{
    char buff[256],*p,sum;
    double h=0.0,ep[6],pos[3],dms1[3],dms2[3];
    
    time2epoch(gpst2utc(t),ep);
    ecef2pos(rr,pos);
    h=geoidh(pos);
    todms(fabs(pos[0])*R2D,dms1); todms(fabs(pos[1])*R2D,dms2);
    sprintf(buff,"$GPGGA,%02.0f%02.0f%05.2f,%02.0f%10.7f,%s,%03.0f%10.7f,%s,%d,%02d,,%.3f,M,%.3f,M,,,",
            ep[3],ep[4],ep[5],dms1[0],dms1[1]+dms1[2]/60.0,pos[0]>=0?"N":"S",
            dms2[0],dms2[1]+dms2[2]/60.0,pos[1]>=0?"E":"W",stat<=2?2:1,ns,
            pos[2]-h,h);
    for (p=buff+1,sum=0;*p;p++) sum^=*p; /* check-sum */
    fprintf(fp,"%s*%02X%c%c",buff,sum,0x0D,0x0A);
}
/* input observation data/navigation messages for an epoch -------------------*/
static int input(obsd_t **obs, nav_t **nav)
{
    int i,n;
    if (!revs) { /* forward */
        for (i=iobs;iobs<obss.n;iobs++) {
            if (timediff(obss.data[iobs].time,obss.data[i].time)>MAXDT) break;
        }
        *obs=obss.data+i; n=iobs-i;
    }
    else { /* backward */
        for (i=iobs;iobs>=0;iobs--) {
            if (timediff(obss.data[i].time,obss.data[iobs].time)>MAXDT) break;
        }
        *obs=obss.data+iobs+1; n=i-iobs;
    }
    *nav=&navs;
    return n;
}
/* output result -------------------------------------------------------------*/
static void output(gtime_t t, sol_t sol, sol_t solf, int stat, int ns)
{
    int week;
    char s[64];
    double gpst,*rr,*Qr;
    
    rr=stat==1?solf.rr:sol.rr; Qr=stat==1?solf.Qr:sol.Qr;
    if (times) t=gpst2utc(t);
    if (timef) time2str(t,s,timeu);
    else {
        gpst=time2gpst(t,&week);
        if (86400*7-gpst<0.5/pow(10.0,timeu)) gpst=0.0;
        sprintf(s,"%*.*f",7+(timeu<=0?0:timeu+1),timeu,gpst);
    }
    switch (posf) {
        case 1:  outecef(fp,s,rr,Qr,stat,ns); break;
        case 2:  outnmea(fp,t,rr,stat,ns);    break;
        default: outpos(fp,s,rr,Qr,stat,ns);  break;
    }
}
/* save result ---------------------------------------------------------------*/
static void saveres(gtime_t t, sol_t sol, sol_t solf, int stat, int ns)
{
    result_t *p;
    if ((revs?iresb:iresf)>=nepoch) return;
    p=revs?resb+iresb++:resf+iresf++;
    p->t=t; p->sol=sol; p->solf=solf; p->stat=stat; p->ns=ns;
}
/* smoothed solution ---------------------------------------------------------*/
static int smoothsol(sol_t solf, sol_t solb, sol_t *sols)
{
    int i;
    double rr[3];
    
    /* Qr=(Qf^-1+Qb^-1)^-1,rr=Qr*(Qf^-1*rf+Qb^-1*rb) */
    if (matinv(solf.Qr,3)||matinv(solb.Qr,3)) return -1;
    for (i=0;i<9;i++) sols->Qr[i]=solf.Qr[i]+solb.Qr[i];
    if (matinv(sols->Qr,3)) return -1;
    matmul("NN",3,1,3,1.0,solf.Qr,solf.rr,0.0,rr);
    matmul("NN",3,1,3,1.0,solb.Qr,solb.rr,1.0,rr);
    matmul("NN",3,1,3,1.0,sols->Qr,rr,0.0,sols->rr);
    return 0;
}
/* combine forward/backward solutions and output results ---------------------*/
static void combres(void)
{
    int i,j,k,stat;
    double tt,dr;
    sol_t sol,solf;
    
    for (i=0,j=iresb-1;i<iresf&&j>=0;i++,j--) {
        tt=timediff(resf[i].t,resb[j].t);
        if (tt<-MAXDT) j++; else if (tt>MAXDT) i--;
        if (tt<-MAXDT||resf[i].stat<resb[j].stat) {
            output(resf[i].t,resf[i].sol,resf[i].solf,resf[i].stat,resf[i].ns);
            continue;
        }
        else if (tt>MAXDT||resf[i].stat>resb[j].stat) {
            output(resb[j].t,resb[j].sol,resb[j].solf,resb[j].stat,resb[j].ns);
            continue;
        }
        /* validation of forward/backward fixed solutions */
        if ((stat=resf[i].stat)==1) {
            for (k=0;k<3;k++) {
                dr=resf[i].solf.rr[k]-resb[j].solf.rr[k];
                if (dr*dr>16.0*(resf[i].solf.Qr[k+k*3]+resb[j].solf.Qr[k+k*3]))
                    stat=2; /* degrade to float solution */
            }
        }
        if (stat==1) { /* fixed */
            if (smoothsol(resf[i].solf,resb[j].solf,&solf)) continue;
        }
        else { /* float,dgps or single */
            if (smoothsol(resf[i].sol,resb[j].sol,&sol)) continue;
        }
        output(timeadd(resf[i].t,-tt/2.0),sol,solf,stat,resf[i].ns);
    }
}
/* rnx2rtkp main -------------------------------------------------------------*/
int main(int argc, char **argv)
{
    int i,j,n;
    char *infile[MAXFILE],*outfile="";
    
    for (i=1,n=0;i<argc;i++) {
        if      (!strcmp(argv[i],"-o")&&i+1<argc) outfile=argv[++i];
        else if (!strcmp(argv[i],"-p")&&i+1<argc) mode=atoi(argv[++i]);
        else if (!strcmp(argv[i],"-f")&&i+1<argc) nfreq=atoi(argv[++i]);
        else if (!strcmp(argv[i],"-m")&&i+1<argc) elmin=atof(argv[++i]);
        else if (!strcmp(argv[i],"-v")&&i+1<argc) thres=atof(argv[++i]);
        else if (!strcmp(argv[i],"-s")&&i+1<argc) sep=argv[++i];
        else if (!strcmp(argv[i],"-d")&&i+1<argc) timeu=atoi(argv[++i]);
        else if (!strcmp(argv[i],"-b")) dirs=1;
        else if (!strcmp(argv[i],"-c")) dirs=2;
        else if (!strcmp(argv[i],"-i")) opt|=1;
        else if (!strcmp(argv[i],"-u")) times=1;
        else if (!strcmp(argv[i],"-t")) timef=1;
        else if (!strcmp(argv[i],"-e")) posf=1;
        else if (!strcmp(argv[i],"-n")) posf=2;
        else if (!strcmp(argv[i],"-g")) degf=1;
        else if (!strcmp(argv[i],"-r")&&i+3<argc)
            for (j=0;j<3;j++) rb[j]=atof(argv[++i]);
        else if (*argv[i]=='-') printhelp();
        else if (n<MAXFILE) infile[n++]=argv[i];
    }
    if (mode<0||3<mode||nfreq<1||2<nfreq||elmin<0.0||90.0<elmin||12<timeu) {
        fprintf(stderr,"%s : command option error\n",__FILE__);
        return -1;
    }
    if (n<=0) {
        fprintf(stderr,"%s : no input file\n",__FILE__);
        return -2;
    }
    /* read rinex obs/nav files */
    if ((nepoch=readrnx(infile,n,&obss,&navs))<=0) {
        fprintf(stderr,"%s : no observation data\n",__FILE__);
        return -3;
    }
    if (navs.n<=0) {
        fprintf(stderr,"%s : no navigation message\n",__FILE__);
        return -4;
    }
    /* compute reference station position */
    if (mode>=1&&norm(rb,3)<=0.0&&refpos(rb)) {
        fprintf(stderr,"%s : no reference position\n",__FILE__);
        return -5;
    }
    if (!*outfile) fp=stdout;
    else if (!(fp=fopen(outfile,"w"))) {
        fprintf(stderr,"%s : output file open error : %s\n",__FILE__,outfile);
        return -6;
    }
    /* output header */
    if (posf<2) outhead(infile,n);
    
    if (dirs<2) { /* forward/backward solution */
        if (dirs) {revs=1; iobs=obss.n-1;}
        rtkpos(input,output,mode,nfreq,rb,elmin*D2R,thres,opt);
    }
    else { /* combined solution */
        resf=(result_t *)malloc(sizeof(result_t)*nepoch);
        resb=(result_t *)malloc(sizeof(result_t)*nepoch);
        if (resf&&resb) {
            rtkpos(input,saveres,mode,nfreq,rb,elmin*D2R,thres,opt);
            revs=1; iobs=obss.n-1;
            rtkpos(input,saveres,mode,nfreq,rb,elmin*D2R,thres,opt);
            
            /* combine forward/backward solutions and output results */
            combres();
        }
        else fprintf(stderr,"%s : memory allocation error\n",__FILE__);
        free(resf); free(resb);
    }
    fclose(fp);
    return 0;
}
