/*------------------------------------------------------------------------------
* rtkpos.c : realtime kinematic positioning
*
*          Copyright (C) 2007 by T.TAKASU, All rights reserved.
*
* version : $Revision: 1.11 $ $Date: 2007/01/23 16:28:05 $
* history : 2007/01/12 1.0 new
*-----------------------------------------------------------------------------*/
#include "rtklib.h"

static const char rcsid[]="$Id: rtkpos.c,v 1.11 2007/01/23 16:28:05 ttaka Exp TTAKA $";

/* constants/macros ----------------------------------------------------------*/

#define NR          3           /* number of real(float) parameters */
#define VARB        100.0       /* initial phase-bias variance (cycle^2) */
#define ERRA        0.004       /* measurement error factor a of phase */
#define ERRB        0.003       /* measurement error factor b of phase */
#define ERRC        0.010       /* measurement error factor c of phase */
#define THRESSLIP   0.05        /* slip threshold of geometry-free phase (m) */
#define MAXINNO     30.0        /* reject threshold of innovation (m) */
#define MAXOUT      5           /* obs outage count to reset bias */
#define MINLOCK     5           /* min lock count to fix ambiguity */

#define IB(s,f)     (NR+(s)-1+MAXSAT*(f)) /* state index of phase-bias */

const static double lam[]={CLIGHT/FREQ1,CLIGHT/FREQ2}; /* carrier wave length (m) */

/* types --------------------------------------------------------------------*/

typedef struct {                /* ambiguity control struct */
    int lock[MAXSAT][NFREQ];    /* lock counter of phase */
    int out[MAXSAT][NFREQ];     /* obs outage counter of phase */
    double gf[MAXSAT];          /* geometry-free phase (m) */
} ambctr_t;

/* single-differenced observable (f=0:L1,1:L2,2:C1,3:P2) ---------------------*/
static double sdobs(const obsd_t *obs, int i, int j, int f)
{
    double pi=f<NFREQ?lam[f]*obs[i].L[f]:obs[i].P[f-NFREQ];
    double pj=f<NFREQ?lam[f]*obs[j].L[f]:obs[j].P[f-NFREQ];
    return pi==0.0||pj==0.0?0.0:pi-pj;
}
/* single-differenced geometry-free linear combination of phase --------------*/
static double gfobs(const obsd_t *obs, int i, int j)
{
    double pi=sdobs(obs,i,j,0),pj=sdobs(obs,i,j,1);
    return pi==0.0||pj==0.0?0.0:pi-pj;
}
/* single-differenced measurement error variance (f=0:L1,1:L2,2:C1,3:P2) -----*/
static double varerr(double el, double bl, int f)
{
    double a2=ERRA*ERRA,b2=ERRB*ERRB,c2=ERRC*ERRC,sinel=sin(el);
    return 2.0*(a2+b2/sinel/sinel+c2*bl*bl*1E-8)*(f<NFREQ?1.0:1E4);
}
/* select common satellites between rover and reference station --------------*/
static int selsats(const obsd_t *obs, const double *azel, int nu, int nr,
                   double elmin, int *sat, int *iu, int *ir)
{
    int i,j,k=0;
    for (i=0,j=nu;i<nu&&j<nu+nr;i++,j++) {
        if      (obs[i].sat<obs[j].sat) j--;
        else if (obs[i].sat>obs[j].sat) i--;
        else if (azel[1+i*2]>=elmin) {sat[k]=obs[i].sat; iu[k]=i; ir[k++]=j;}
    }
    return k;
}
/* initialize state and covariance -------------------------------------------*/
static void initx(double *x, double *P, int nx, double xi, double var, int i)
{
    int j;
    x[i]=xi; for (j=0;j<nx;j++) P[i+j*nx]=P[j+i*nx]=i==j?var:0.0;
}
/* temporal update of states -------------------------------------------------*/
static void udstate(double *x, double *P, int nx, int nf, const obsd_t *obs,
                    const int *sat, const int *iu, const int *ir, int ns,
                    ambctr_t *ambc, int opt)
{
    int i,j,f,reset,slip[MAXSAT]={0};
    double g0,g1,cp,pr,bias[MAXSAT]={0.0},offset;
    
    for (i=0;i<ns;i++) {
        /* detect slip by geometry-free phase-jump */
        if ((g1=gfobs(obs,iu[i],ir[i]))==0) continue;
        g0=ambc->gf[sat[i]-1]; ambc->gf[sat[i]-1]=g1;
        slip[i]=g0!=0.0&&fabs(g1-g0)>THRESSLIP;
    }
    for (f=0;f<nf;f++) {
        /* reset phase-bias if instantaneous AR or expire obs outage counter */
        for (i=1;i<=MAXSAT;i++) {
            reset=++ambc->out[i-1][f]>MAXOUT;
            if ((opt||reset)&&x[IB(i,f)]!=0.0) initx(x,P,nx,0.0,0.0,IB(i,f));
            if (!opt&&reset) ambc->lock[i-1][f]=-MINLOCK;
        }
        /* reset phase-bias if slip */
        for (i=0;i<ns;i++) {
            if (opt||!slip[i]) continue;
            x[IB(sat[i],f)]=0.0; ambc->lock[sat[i]-1][f]=-MINLOCK;
        }
        for (i=j=0,offset=0.0;i<ns;i++) {
            /* approx single-differenced phase-bias by phase - pseudorange */
            cp=sdobs(obs,iu[i],ir[i],f); pr=sdobs(obs,iu[i],ir[i],NFREQ);
            if (cp!=0.0&&pr!=0.0) bias[i]=(cp-pr)/lam[f]; 
            if (x[IB(sat[i],f)]!=0.0) {offset+=bias[i]-x[IB(sat[i],f)]; j++;}
        }
        /* correct phase-bias offset caused by receiver clock-jump */
        if (j>0) {
            for (i=1;i<=MAXSAT;i++) if (x[IB(i,f)]!=0.0) x[IB(i,f)]+=offset/j;
        }
        /* set initial phase-bias */
        for (i=0;i<ns;i++) {
            if (bias[i]==0.0||x[IB(sat[i],f)]!=0.0) continue;
            initx(x,P,nx,bias[i],VARB,IB(sat[i],f));
        }
    }
}
/* double-differenced phase/code residuals -----------------------------------*/
static int residual(double *x, int nx, int nf, const obsd_t *obs,
                    const int *sat, const int *iu, const int *ir, int ns,
                    const double *rs, const double *azel, const double *rb,
                    int mode, double *v, double *H, double *var, int *n, int *vs)
{
    int i,j,k,f,nv=0;
    double du,dr,yi=0.0,yj,bias,bl[3],posu[3],posr[3],azelr[2];
    double eu[3*MAXSAT],er[3],d[MAXSAT]={0.0};
    
    ecef2pos(x,posu); ecef2pos(rb,posr); for (i=0;i<3;i++) bl[i]=x[i]-rb[i];
    for (i=0;i<ns;i++) {
        /* single-differenced geometric distance + tropo-model */
        du=geodist(rs+iu[i]*3,x,eu+i*3); dr=geodist(rs+ir[i]*3,rb,er);
        if (du<=0.0||dr<=0.0) continue;
        satazel(posr,er,azelr);
        d[i]=du-dr+tropmodel(posu,azel+iu[i]*2)-tropmodel(posr,azelr);
    }
    for (f=mode>=2?0:NFREQ;f<NFREQ*2;f+=NFREQ-nf+1) {
        /* search valid observable and make double-difference */
        for (i=0;i<ns;i++) {
            if (d[i]!=0.0&&(yi=sdobs(obs,iu[i],ir[i],f))!=0.0) break;
        }
        for (j=i+1;j<ns;j++) {
            if (d[j]==0.0||(yj=sdobs(obs,iu[j],ir[j],f))==0.0) continue;
            
            bias=f<NFREQ?lam[f]*(x[IB(sat[i],f)]-x[IB(sat[j],f)]):0.0;
            
            /* residual (measurement - model) */
            if (fabs(v[nv]=(yi-yj)-(d[i]-d[j]+bias))>MAXINNO) continue;
            
            /* partial derivatives */
            for (k=0;k<3;k++) H[k+nv*nx]=-eu[k+i*3]+eu[k+j*3];
            if (f<NFREQ) {
                H[IB(sat[i],f)+nv*nx]=lam[f]; H[IB(sat[j],f)+nv*nx]=-lam[f];
                vs[f+i*NFREQ]=vs[f+j*NFREQ]=1;
            }
            /* measurement error variance */
            var[  nv*2]=varerr(azel[1+iu[i]*2],norm(bl,3),f);
            var[1+nv*2]=varerr(azel[1+iu[j]*2],norm(bl,3),f);
            nv++; n[f]++;
        }
    }
    return nv; /* return number of v */
}
/* double-differenced measurement error covariance ---------------------------*/
static void coverr(int nf, int nv, const int *n, const double *var, double *R)
{
    int i,j,k,f;
    for (f=i=0;f<NFREQ*2;f+=NFREQ-nf+1,i=j) {
        for (j=i;j<i+n[f];j++) {
            for (k=j;k<i+n[f];k++) R[j+k*nv]=R[k+j*nv]=var[j*2];
            R[j+j*nv]+=var[1+j*2];
        }
    }
}
/* relative positioning ------------------------------------------------------*/
static int relpos(double *x, double *P, int nx, int nf, const obsd_t *obs,
                  int nu, int nr, const double *rs, const double *azel,
                  ambctr_t *ambc, const double *rb, double elmin, int mode,
                  int opt)
{
    int i,f,ns,ny,nv,info=-1;
    int n[NFREQ*2]={0},sat[MAXSAT],iu[MAXSAT],ir[MAXSAT],vs[NFREQ*MAXSAT]={0};
    double s,*v,*H,*var,*R,*xp=mat(nx,1),*Pp=mat(nx,nx);
    
    /* select common satellites */
    ns=selsats(obs,azel,nu,nr,elmin,sat,iu,ir);
    
    /* temporal update of states */
    udstate(x,P,nx,nf,obs,sat,iu,ir,ns,ambc,opt);
    
    ny=ns*nf*NFREQ; v=mat(ny,1); H=zeros(nx,ny); var=mat(2,ny);
    
    /* double-differenced phase/code residuals (innovation) */
    nv=residual(x,nx,nf,obs,sat,iu,ir,ns,rs,azel,rb,mode,v,H,var,n,vs);
    
    if ((R=zeros(nv,nv))) {
        /* measurement error covairance */
        coverr(nf,nv,n,var,R);
        
        /* kalman filter state update */
        info=filter(x,P,H,v,R,nx,nv,xp,Pp);
        
        /* postfit residual test */
        nv=residual(xp,nx,nf,obs,sat,iu,ir,ns,rs,azel,rb,mode,v,H,var,n,vs);
        for (i=0,s=0.0;i<nv;i++) s+=v[i]*v[i]/(var[i*2]+var[1+i*2]);
        if (nv>NR&&s>chisqr[nv-NR-1]) info=-1;
        
        if (!info) {
            /* update ambiguity control struct */
            for (f=0;f<nf;f++) {
                for (i=0;i<ns;i++) {
                    if (!vs[f+i*NFREQ]) continue;
                    ambc->lock[sat[i]-1][f]++; ambc->out[sat[i]-1][f]=0;
                }
            }
            memcpy(x,xp,sizeof(double)*nx); memcpy(P,Pp,sizeof(double)*nx*nx);
        }
    }
    free(xp); free(Pp); free(v); free(H); free(var); free(R);
    return info;
}
/* resolve integer ambiguity -------------------------------------------------*/
static int resamb(double *x, double *P, int nx, int nf, double thres,
                  ambctr_t *ambc, double *rr, double *Qr)
{
    int i,j,k,f,ny,nb=0,info;
    double *D,*DP,*y,*Qy,*b,*Qb,*Qrb,*QQ,s[2];
    
    /* single to double-difference transformation matrix (D') */
    D=zeros(nx,nx); for (i=0;i<NR;i++) D[i+i*nx]=1.0;
    for (f=0,k=NR;f<nf;f++,k+=MAXSAT) {
        for (i=k  ;i<k+MAXSAT;i++) if (x[i]!=0.0&&ambc->lock[i-k][f]>0) break;
        for (j=i+1;j<k+MAXSAT;j++) if (x[j]!=0.0&&ambc->lock[j-k][f]>0) {
            D[i+(NR+nb)*nx]=1.0; D[j+(NR+nb++)*nx]=-1.0;
        }
    }
    if (nb<=0) {free(D); return -1;} /* no valid double-difference */
    
    /* transform single to double-differenced phase-bias (y=D'*x, Qy=D'*P*D) */
    ny=NR+nb; y=mat(ny,1); Qy=mat(ny,ny); DP=mat(ny,nx);
    matmul("TN",ny,1,nx,1.0,D,x,0.0,y);
    matmul("TN",ny,nx,nx,1.0,D,P,0.0,DP);
    matmul("NN",ny,ny,nx,1.0,DP,D,0.0,Qy);
    free(D); free(DP);
    
    /* phase-bias covariance (Qb) and real-parameters to bias covariance (Qrb) */
    b=mat(nb,2); Qb=mat(nb,nb); Qrb=mat(NR,nb); QQ=mat(NR,nb);
    for (i=0;i<nb;i++) for (j=0;j<nb;j++) Qb[i+j*nb]=Qy[NR+i+(NR+j)*ny];
    for (i=0;i<NR;i++) for (j=0;j<nb;j++) Qrb[i+j*NR]=Qy[i+(NR+j)*ny];
    
    /* lambda/mlambda integer least-square estimation */
    if (!(info=lambda(nb,2,y+NR,Qb,b,s))) {
        
        /* validation by popular ratio-test */
        if (s[1]/s[0]>=thres) {
            
            /* transform float to fixed solution (rr=rr-Qrb*Qb\(b0-b)) */
            for (i=0;i<nb;i++) y[NR+i]-=b[i];
            if (!(info=matinv(Qb,nb))) {
                matmul("NN",nb,1,nb,1.0,Qb,y+NR,0.0,b);
                matmul("NN",NR,1,nb,-1.0,Qrb,b,1.0,rr);
                
                /* covariance (Qr=Qr-Qrb*Qb^-1*Qrb') */
                matmul("NN",NR,nb,nb,1.0,Qrb,Qb,0.0,QQ);
                matmul("NT",NR,NR,nb,-1.0,QQ,Qrb,1.0,Qr);
            }
        }
        else info=-1; /* validation failed */
    }
    free(y); free(Qy); free(b); free(Qb); free(Qrb); free(QQ);
    return info; /* 0:ok */
}
/* relative positioning --------------------------------------------------------
* input observation data and navigation message, compute rover position by 
* relative positioning and output results. if callback input() return 0, exit
* the function.
* args   : infunc_t input   I   input callback function
*          outfunc_t output I   output callback function
*          int    mode      I   mode (0:single,1:dgps,2:kinematic,3:static)
*          int    nf        I   number of frequencies (1:L1,2:L1+L2)
*          double *rb       I   reference (base) station position (ecef) (m)
*          double elmin     I   elevation mask angle (rad)
*          double thres     I   validation threshold of ambiguity (1>:no AR)
*          int    opt       I   options (1:instantaneous AR)
* return : none
* notes  : input and output callback shall comply the following prototypes.
*
*          int input(obsd_t *obs, nav_t *nav);
*            obsd_t *obs    I   observation data records for an epoch
*                               obs[i].rcv=1:rover,2:reference
*                               sorted by receiver and satellte
*            nav_t  *nav    I   navigation messages
*            return             number of observation data records
*
*          void output(gtime_t t, sol_t sol, sol_t solf, int stat, int ns);
*            gtime_t t      I   time
*            sol_t sol,solf I   float, dgps or single solution, fixed solution
*            int    stat    I   status (1:fix,2:float,4:dgps,5:single)
*            int    ns      I   number of valid satellites
*-----------------------------------------------------------------------------*/
extern void rtkpos(infunc_t input, outfunc_t output, int mode, int nf,
                   const double *rb, double elmin, double thres, int opt)
{
    int i,j,nx=NR+MAXSAT*nf,ny,nu,nr,ns,stat;
    double rs[3*MAXOBS],dts[MAXOBS],azel[2*MAXOBS],dtr,*x,*P;
    ambctr_t ambc={{{0}}};
    obsd_t *obs;
    nav_t *nav;
    sol_t sol,solf;
    
    x=zeros(nx,1); P=zeros(nx,nx);
    
    /* input observation data and navigation messages for an epoch */
    while ((ny=input(&obs,&nav))>0) {
        
        /* count rover/reference station observations */
        for (nu=0;nu<ny&&obs[nu].rcv==1;nu++) ;
        for (nr=0;nu+nr<ny&&obs[nu+nr].rcv==2;nr++) ;
        
        /* satellite positions/clock-biases at signal transmission */
        satpos(obs,ny,nav,rs,dts);
        
        /* rover position by single point positioning */
        if ((ns=pntpos(obs,nu,rs,dts,nav->ion,elmin,sol.rr,sol.Qr,&dtr,azel))<0)
            continue;
        stat=5;
        
        if (mode>=1) {
            if (mode<3||norm(x,3)<=0.0) { /* kinematic mode or first epoch? */
                for (i=0;i<3;i++) {
                    x[i]=sol.rr[i];
                    for (j=i;j<nx;j++) P[i+j*nx]=P[j+i*nx]=j<3?sol.Qr[i+j*3]:0.0;
                }
            }
            /* rover position by relative positioning */
            if (!relpos(x,P,nx,nf,obs,nu,nr,rs,azel,&ambc,rb,elmin,mode,opt)) {
                
                for (i=0;i<3;i++) {
                    sol.rr[i]=x[i]; for (j=0;j<3;j++) sol.Qr[i+j*3]=P[i+j*nx];
                }
                solf=sol; stat=mode<=1?4:2; /* 1:fix,2:float,4:dgps,5:single */
                
                /* resolve integer ambiguity */
                if (mode>=2&&thres>=1.0) {
                    if (!resamb(x,P,nx,nf,thres,&ambc,solf.rr,solf.Qr)) stat=1;
                }
            }
        }
        /* output results */
        output(timeadd(obs[0].time,-dtr),sol,solf,stat,ns);
    }
    free(x); free(P);
}
